/*
 * Copyright (c) 2001 Massachusetts Institute of Technology.
 *
 * This file is part of bgptools. 
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `COPYING' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file AUTHORS which should have accompanied this software 
 * for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the AUTHORS file.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


/*
 *     Nick Feamster (feamster@lcs.mit.edu)
 *     MIT Lab for Computer Science
 *     November 2001
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <math.h>

#include "common.h"
#include "bgp_dump.h"
#include "config.h"


const char *
intoa(u_int32_t addr)
{
	register char *cp;
	register u_int byte;
	register int n;
	static char buf[sizeof(".xxx.xxx.xxx.xxx")];

	cp = &buf[sizeof buf];
	*--cp = '\0';

	n = 4;
	do {
		byte = addr & 0xff;
		*--cp = byte % 10 + '0';
		byte /= 10;
		if (byte > 0) {
			*--cp = byte % 10 + '0';
			byte /= 10;
			if (byte > 0)
				*--cp = byte + '0';
		}
		*--cp = '.';
		addr >>= 8;
	} while (--n > 0);

	return cp + 1;
}

void read_record(FILE *fp, update *upd, table_entry *te) {
 
  u_int32_t tmpbuf=0;
  char tmpbuf_l[16];
  u_int16_t subtype=0, src_as=0, dst_as=0, if_ind=0, af=0;
  u_int32_t len=0, src_ip=0, dst_ip=0;

  /* specific to BGP messages */
  u_int16_t msg_length=0, msg_type=0, wr_len=0, pa_len=0, nlri_len=0;
  char *tmpbuf_var = NULL;

  int count=0, i=0;


  /* BGP update attributes */
  char origin=0;
  char as_path[1024*200];
  char *next_hop_str;
  u_int32_t next_hop=0, med=0, localpref=0;
  char aggregator[6];

  char pfxlen, pa_fgs, pa_type;
  char pa_fgs_str[4];
  u_int16_t attr_len=0;
  
  u_int32_t pfx=0;
  char *pfxstr;

  char ps_type, ps_len;
  u_int16_t as_el=0;

  /* for table dumps */
  upd->prefix_=NULL;
  upd->prefix_wd_=NULL;
  upd->num_prefixes=0;
  upd->num_prefixes_wd=0;
    
/*    fprintf(stderr, "reading record...\n"); */

  while (!feof(fp)) {
    fread((void*)&tmpbuf, sizeof(char), 4, fp);
    tmpbuf = ntohl(tmpbuf);
    
    if (tmpbuf == ((MSG_PROTOCOL_BGP4MP<<16)|BGP4MP_MESSAGE) ||
	tmpbuf>>16 == MSG_TABLE_DUMP) {
      fseek(fp, -8, SEEK_CUR);
      break;
    } else {
	if (!feof(fp))
	  fseek(fp, -3, SEEK_CUR);
    }
  }

  /* read first 4 bytes -- should be time */
  fread((void*)&tmpbuf, sizeof(char), 4, fp);
  tmpbuf = ntohl(tmpbuf);

  upd->time = (u_int32_t)tmpbuf;

#ifdef BGPDUMP_DEBUG
  fprintf(stdout, "\n**************************\n");
  fprintf(stdout, "time is: %d\n", upd->time);
#endif

  /* type and subtype */
  fread((void*)&tmpbuf, sizeof(char), 4, fp);
  tmpbuf = ntohl(tmpbuf);
  upd->dump_type = (u_int16_t)(tmpbuf>>16);
  subtype = (u_int16_t)(tmpbuf&65535);

  switch(upd->dump_type) {
  case MSG_PROTOCOL_BGP4MP:
    switch(subtype) {

    case BGP4MP_STATE_CHANGE:
      break;

    case BGP4MP_MESSAGE:

#ifdef BGPDUMP_DEBUG
      fprintf(stdout, "type is BGP4MP/Message\n");
#endif

      /* not sure what this field is all about... */
      fread((void*)&tmpbuf, sizeof(char), 4, fp);
      tmpbuf = ntohl(tmpbuf);
      len = tmpbuf;

      /* read src AS and dest. AS */
      fread((void*)&tmpbuf, sizeof(char), 4, fp);
      tmpbuf = ntohl(tmpbuf);
      src_as = (u_int16_t)(tmpbuf>>16);
      dst_as = (u_int16_t)(tmpbuf&65535);

      upd->src_as = src_as;
      upd->dst_as = dst_as;

#ifdef BGPDUMP_DEBUG
      fprintf(stdout, "src AS: %d, dest AS: %d\n", src_as, dst_as);
#endif

      /* interface id and address family */
      fread((void*)&tmpbuf, sizeof(char), 4, fp);
      tmpbuf = ntohl(tmpbuf);
      if_ind = (u_int16_t)(tmpbuf>>16);
      af = (u_int16_t)(tmpbuf&65535);

      upd->if_ind = if_ind;
      upd->addr_fam = af;

#ifdef BGPDUMP_DEBUG
      fprintf(stdout, "if #: %d, af: %d\n", if_ind, af);
#endif
      /* src IP address */
      fread((void*)&tmpbuf, sizeof(char), 4, fp);
      tmpbuf = ntohl(tmpbuf);
      src_ip = (u_int32_t)(tmpbuf);

      /* destination IP address */
      fread((void*)&tmpbuf, sizeof(char), 4, fp);
      tmpbuf = ntohl(tmpbuf);
      dst_ip = (u_int32_t)(tmpbuf);

      /**** read contents of BGP Message ****/

      /* BGP HEADER -- typically all ones so we ignore */
      fread((void*)&tmpbuf_l, sizeof(char), 16, fp);

      /* BGP message length and type */
      fread((void*)&msg_length, sizeof(char), 2, fp);
      msg_length = ntohs(msg_length);
      fread((void*)&msg_type, sizeof(char), 1, fp);
      msg_type &= 0xff;

      upd->msg_length = msg_length;
      upd->msg_type = msg_type;

#ifdef BGPDUMP_DEBUG
      fprintf(stdout, "message length: %d, type: %d\n", msg_length, msg_type);
#endif

      switch(msg_type) {
      case BGP_OPEN:
	break;
      case BGP_UPDATE:
#ifdef BGPDUMP_DEBUG
	fprintf(stdout, "UPDATE Message Received...\n");
#endif
	/* read length of withdrawn routes */
	fread((void*)&wr_len, sizeof(char), 2, fp);
	wr_len = ntohs(wr_len);
#ifdef BGPDUMP_DEBUG
	fprintf(stdout, "withdrawn routes length: %d\n", wr_len);
#endif

	if (wr_len) {
	  upd->ad_type = BGP_WITHDRAWAL;
	  tmpbuf_var = (char*)realloc((void*)tmpbuf_var, wr_len*sizeof(char));
	  fread((void*)tmpbuf_var, sizeof(char), wr_len, fp);
	  /* read out withdrawn routes */
	  count = 0;

	  while (count < wr_len) {
  	    upd->num_prefixes_wd++;
	    memcpy((void*)&pfxlen, (void*)(tmpbuf_var+count), 1*sizeof(char));
	    count++;
  	    pfxlen &= 0xff; 
	    
	    if ((u_int16_t)pfxlen >32) {
	      break;
	    }

	    memcpy((void*)&pfx, (void*)(tmpbuf_var+count),
		 (pfxlen/8 + ((pfxlen%8)?1:0))*sizeof(char));
	    pfx = ntohl(pfx);
	    pfx &= ((unsigned int)pow(2,pfxlen)-1) << (32-pfxlen);
	    pfxstr = (char*)intoa(pfx);


	    // WEIRD MEMORT SHIT REQUIRES THIS
  	    if (upd->prefix_==NULL || (upd->num_prefixes*sizeof(prefix_s) > sizeof(*(upd->prefix_))))
  	      upd->prefix_ = (prefix_s *)realloc((void*)(upd->prefix_), 2*upd->num_prefixes*sizeof(prefix_s));

	    if (upd->prefix_wd_==NULL || (upd->num_prefixes_wd*sizeof(prefix_s) > sizeof(*(upd->prefix_wd_))))
	      upd->prefix_wd_ = (prefix_s *)realloc((void*)(upd->prefix_wd_), 2*upd->num_prefixes_wd*sizeof(prefix_s));
    	    (upd->prefix_wd_)[upd->num_prefixes_wd-1].pfx = pfx;  
    	    (upd->prefix_wd_)[upd->num_prefixes_wd-1].len = pfxlen;  



	    
#ifdef BGPDUMP_DEBUG
	    fprintf(stdout, "prefix: 0x%.4x (%s), length: %d\n", pfx, pfxstr, pfxlen);
#endif
	    count+=(pfxlen/8) + ((pfxlen%8)?1:0);
	  }
	}


	/* path attribute length) */
	fread((void*)&pa_len, sizeof(char), 2, fp);
	pa_len = ntohs(pa_len);
#ifdef BGPDUMP_DEBUG
	fprintf(stdout, "path attr. length: %d\n", pa_len);
#endif
	if (pa_len) {
	  upd->ad_type = BGP_ADV;
	  tmpbuf_var = (char*)realloc((void*)tmpbuf_var, pa_len*sizeof(char));
	  fread((void*)tmpbuf_var, sizeof(char), pa_len, fp);

	  /* read out all of the path attribute stuff */
	  count = 0;

	  while (count < pa_len) {

	    /* parse attribute flags */
	    memcpy((void*)&pa_fgs, (void*)(tmpbuf_var+count), 1*sizeof(char));
	    count++;

	    pa_fgs_str[0] = '\0';
#ifdef BGPDUMP_DEBUG
	    if (pa_fgs&0x80) {
	      strncat(pa_fgs_str, "O", 1);
	    }
	    if (pa_fgs&0x40) {
	      strncat(pa_fgs_str, "T", 1);
	    }
	    if (pa_fgs&0x20) {
	      strncat(pa_fgs_str, "P", 1);
	    }
	    if (pa_fgs&0x10) {
	      strncat(pa_fgs_str, "E", 1);
	    }
	    if (strlen(pa_fgs_str) > 0) { fprintf(stdout, "(%s)\t", pa_fgs_str); }
#endif

	    /* parse attribute type */
	    memcpy((void*)&pa_type, (void*)(tmpbuf_var+count), 1*sizeof(char));
	    count++;

	    /* parse attribute length */
	    if (pa_fgs&0x10) {
	      memcpy((void*)&attr_len, (void*)(tmpbuf_var+count), 2*sizeof(char));
	      attr_len = ntohs(attr_len);
	      count += 2;
	    } else {
	      memcpy((void*)&attr_len, (void*)(tmpbuf_var+count), 1*sizeof(char));
	      count += 1;
	      attr_len &= 0xff;
	    }  

	    /* parse attribute value */
	    as_path[0] = '\0';

	    switch(pa_type) {
	    case ORIGIN:
	      memcpy((void*)&origin, (tmpbuf_var+count), attr_len*sizeof(char));
	      origin &= 0x3;
	      upd->origin = origin;
#ifdef BGPDUMP_DEBUG
	      fprintf(stdout, "\tORIGIN: %d\n", origin);
#endif
	      break;
	    case AS_PATH:
	      memcpy((void*)&ps_type, (tmpbuf_var+count), 1*sizeof(char));
	      memcpy((void*)&ps_len, (tmpbuf_var+count+1), 1*sizeof(char));
	      for (i=0; i < 2*ps_len; i+=2) {
		memcpy((void*)&as_el, (tmpbuf_var+count+2+i), 2*sizeof(char));
		as_el = ntohs(as_el);
		sprintf(as_path, "%s%d ", as_path, as_el);
		if (strlen(as_path) > 256) {break;}
	      }
	      upd->origin_as = as_el;

	      as_path[strlen(as_path)-1] = '\0';
  	      strncpy(upd->as_path, as_path, strlen(as_path));
	      upd->as_path[strlen(as_path)] = '\0';
#ifdef BGPDUMP_DEBUG
	      fprintf(stdout, "\tASPATH: %s\n", as_path);
#endif
	      break;
	    case NEXT_HOP:
	      memcpy((void*)&next_hop, (tmpbuf_var+count), attr_len*sizeof(char));
	      next_hop = ntohl(next_hop);
	      next_hop_str = (char*)intoa(next_hop);
	      upd->next_hop = next_hop;
#ifdef BGPDUMP_DEBUG
	      fprintf(stdout, "\tNEXT HOP: %s\n", next_hop_str);
#endif
	      break;
	    case MULTI_EXIT_DISC:
  	      memcpy((void*)&med, (tmpbuf_var+count), attr_len*sizeof(char));
	      med = ntohl(med);
	      upd->med = med;
#ifdef BGPDUMP_DEBUG
	      fprintf(stdout, "\tMED: %d\n", med);
#endif
	      break;
	    case LOCAL_PREF:
	      memcpy((void*)&localpref, (tmpbuf_var+count), attr_len*sizeof(char));
	      localpref = ntohl(localpref);
	      upd->localpref = localpref;
#ifdef BGPDUMP_DEBUG
	      fprintf(stdout, "\tLOCALPREF: %d\n", localpref);
#endif
	      break;
	    case ATOMIC_AGGREGATE:
	      break;
	    case AGGREGATOR:
#ifdef BGPDUMP_DEBUG
	      fprintf(stdout, "\tAGGREGATOR: %s\n", "XXX");
#endif
	      break;
	    }
	    count += attr_len;
	  }

	  nlri_len = msg_length - 23 - pa_len - wr_len;
	  if (nlri_len && pa_len) {
	    /* read out advertised routes */ 
	    count=0;
	    tmpbuf_var = (char*)realloc((void*)tmpbuf_var, nlri_len*sizeof(char));
	    fread((void*)tmpbuf_var, sizeof(char), nlri_len, fp);
	    while (count < nlri_len) {
	    
	      memcpy((void*)&pfxlen, (void*)(tmpbuf_var+count), 1*sizeof(char)); 
	      count++; 
	      pfxlen &= 0xff; 

	      /* XXX hack segfault guard */
	      if ((u_int16_t)pfxlen >32) {
		break;
	      }

 	      memcpy((void*)&pfx, (void*)(tmpbuf_var+count),
 		     (pfxlen/8 + ((pfxlen%8)?1:0))*sizeof(char));
	      count+=(pfxlen/8) + ((pfxlen%8)?1:0);
	      
	      pfx = ntohl(pfx);
	      pfx &= ((unsigned int)pow(2,pfxlen)-1) << (32-pfxlen);
	      pfxstr = (char*)intoa(pfx);


	      upd->num_prefixes++;
	      
	      if (upd->prefix_==NULL || (upd->num_prefixes*sizeof(prefix_s) > sizeof(*(upd->prefix_))))
		upd->prefix_ = (prefix_s *)realloc((void*)(upd->prefix_), 2*upd->num_prefixes*sizeof(prefix_s));

	      
	      if (!(upd->prefix_)) { break; }

	      (upd->prefix_)[upd->num_prefixes-1].pfx = pfx;  
	      (upd->prefix_)[upd->num_prefixes-1].len = pfxlen;

#ifdef BGPDUMP_DEBUG
	      fprintf(stdout, "prefix: 0x%.4x (%s), length: %d\n", pfx, pfxstr, pfxlen);
#endif

	    }
	  }
	}

	break;
      case BGP_NOTIFICATION:
	break;
      case BGP_KEEPALIVE:
	break;
      default: break;
      }

     
    case BGP4MP_ENTRY:
      break;

    case BGP4MP_SNAPSHOT:
      break;
    }
    break;
    

    /* old school MRT table dump format (a la Zebra) */

  case MSG_TABLE_DUMP:
    
    // this part is still a bit busted, so
    // short circuit it for now...
    break;

#ifdef BGPDUMP_DEBUG
    fprintf(stderr, "MRT table dump...\n");
#endif

    fread((void*)&tmpbuf, sizeof(char), 4, fp);
#ifdef BGPDUMP_DEBUG
    fprintf(stderr, "Mystery: %x\n", ntohl(tmpbuf));
#endif

    fread((void*)&tmpbuf, sizeof(char), 4, fp);
    tmpbuf = ntohl(tmpbuf);
    te->view = (tmpbuf>>16)&0xffff;
    te->seq = (tmpbuf)&0xffff;
#ifdef BGPDUMP_DEBUG
    fprintf(stderr, "View: %d, Seq: %d\n", te->view, te->seq);
#endif

    fread((void*)&tmpbuf, sizeof(char), 4, fp);
    tmpbuf = ntohl(tmpbuf);
    strcpy(te->tpfx, intoa(tmpbuf));
#ifdef BGPDUMP_DEBUG
    fprintf(stderr, "Prefix: %s\n", te->tpfx);
#endif

    fread((void*)&tmpbuf, sizeof(char), 2, fp);
    tmpbuf = ntohs((u_int16_t)tmpbuf);
    te->tlen = (tmpbuf)>>8&0xff;
    te->tstat = (tmpbuf)&0xff;

#ifdef BGPDUMP_DEBUG
    fprintf(stderr, "Length: %d\n", te->tlen);
#endif

    fread((void*)&tmpbuf, sizeof(char), 4, fp);
    tmpbuf = ntohl((u_int32_t)tmpbuf);
    te->uptime = tmpbuf;
#ifdef BGPDUMP_DEBUG
    fprintf(stderr, "Uptime: %d\n", te->uptime);
#endif

    fread((void*)&tmpbuf, sizeof(char), 4, fp);
    tmpbuf = ntohl(tmpbuf);
    strcpy(te->peer_ip, intoa(tmpbuf));
#ifdef BGPDUMP_DEBUG
    fprintf(stderr, "Peer IP: %s\n", te->peer_ip);
#endif

    fread((void*)&tmpbuf, sizeof(char), 2, fp);
    tmpbuf = ntohs(tmpbuf);
    te->peer_as = tmpbuf;
#ifdef BGPDUMP_DEBUG
    fprintf(stderr, "Peer AS: %d\n", te->peer_as);
#endif
    break;


  default: break;
  }

  if (tmpbuf_var) { free(tmpbuf_var); }  
}

/*********************************************************************/

void read_record(char *pkt, int plen, update *upd) {
 
  u_int32_t tmpbuf;
  char tmpbuf_l[16];
  u_int16_t subtype, src_as, dst_as, if_ind, af;
  u_int32_t len, src_ip, dst_ip;

  /* specific to BGP messages */
  u_int16_t msg_length, msg_type, wr_len, pa_len, nlri_len;
  char *tmpbuf_var = NULL;

  int count, i;
  int pkt_offset=0;


  /* BGP update attributes */
  char origin;
  char as_path[80];
  char *next_hop_str;
  u_int32_t next_hop, med, localpref;
  char aggregator[6];

  char pfxlen, pa_fgs, pa_type;
  char pa_fgs_str[4];
  u_int16_t attr_len;
  
  u_int32_t pfx;
  char *pfxstr;

  char ps_type, ps_len;
  u_int16_t as_el;

  upd->num_prefixes=0;
  upd->num_prefixes_wd=0;

  
/*    fprintf(stderr, "reading record...\n"); */

  while (pkt_offset < plen) {
    memcpy((void*)&tmpbuf, (char*)(pkt+pkt_offset), 4);
    pkt_offset+=4;

    tmpbuf = ntohl(tmpbuf);
    
    if (tmpbuf == ((MSG_PROTOCOL_BGP4MP<<16)|BGP4MP_MESSAGE)) {
      pkt_offset -= 8;
      break;
    } else {
      if (pkt_offset < plen)
	pkt_offset -= 3;
    }
  }

  /* read first 4 bytes -- should be time */
  memcpy((void*)&tmpbuf, (char*)(pkt+pkt_offset), 4);
  pkt_offset+=4;
  tmpbuf = ntohl(tmpbuf);

  upd->time = (u_int32_t)tmpbuf;

#ifdef BGPDUMP_DEBUG
  fprintf(stdout, "\n**************************\n");
  fprintf(stdout, "time is: %d\n", upd->time);
#endif

  /* type and subtype */
  memcpy((void*)&tmpbuf, (char*)(pkt+pkt_offset), 4);
  pkt_offset+=4;
  tmpbuf = ntohl(tmpbuf);
  upd->dump_type = (u_int16_t)(tmpbuf>>16);
  subtype = (u_int16_t)(tmpbuf&65535);
  
  switch(upd->dump_type) {
  case MSG_PROTOCOL_BGP4MP:
    switch(subtype) {

    case BGP4MP_STATE_CHANGE:
      break;

    case BGP4MP_MESSAGE:

#ifdef BGPDUMP_DEBUG
      fprintf(stdout, "type is BGP4MP/Message\n");
#endif

      /* not sure what this field is all about... */
      memcpy((void*)&tmpbuf, (char*)(pkt+pkt_offset), 4);
      pkt_offset+=4;

      tmpbuf = ntohl(tmpbuf);
      len = tmpbuf;

      /* read src AS and dest. AS */
      memcpy((void*)&tmpbuf, (char*)(pkt+pkt_offset), 4);
      pkt_offset+=4;
      tmpbuf = ntohl(tmpbuf);
      src_as = (u_int16_t)(tmpbuf>>16);
      dst_as = (u_int16_t)(tmpbuf&65535);

      upd->src_as = src_as;
      upd->dst_as = dst_as;

#ifdef BGPDUMP_DEBUG
      fprintf(stdout, "src AS: %d, dest AS: %d\n", src_as, dst_as);
#endif

      /* interface id and address family */
      memcpy((void*)&tmpbuf, (char*)(pkt+pkt_offset), 4);
      pkt_offset+=4;
      tmpbuf = ntohl(tmpbuf);
      if_ind = (u_int16_t)(tmpbuf>>16);
      af = (u_int16_t)(tmpbuf&65535);

      upd->if_ind = if_ind;
      upd->addr_fam = af;

#ifdef BGPDUMP_DEBUG
      fprintf(stdout, "if #: %d, af: %d\n", if_ind, af);
#endif
      /* src IP address */
      memcpy((void*)&tmpbuf, (char*)(pkt+pkt_offset), 4);
      pkt_offset+=4;

      tmpbuf = ntohl(tmpbuf);
      src_ip = (u_int32_t)(tmpbuf);

      /* destination IP address */
      memcpy((void*)&tmpbuf, (char*)(pkt+pkt_offset), 4);
      pkt_offset+=4;

      tmpbuf = ntohl(tmpbuf);
      dst_ip = (u_int32_t)(tmpbuf);

      /**** read contents of BGP Message ****/

      /* BGP HEADER -- typically all ones so we ignore */
      memcpy((void*)&tmpbuf_l, (char*)(pkt+pkt_offset), 16);
      pkt_offset+=16;


      /* BGP message length and type */
      memcpy((void*)&msg_length, (char*)(pkt+pkt_offset), 2);
      pkt_offset += 2;
      msg_length = ntohs(msg_length);

      memcpy((void*)&msg_type, (char*)(pkt+pkt_offset), 1);
      pkt_offset += 1;
      msg_type &= 0xff;

      upd->msg_length = msg_length;
      upd->msg_type = msg_type;

#ifdef BGPDUMP_DEBUG
      fprintf(stdout, "message length: %d, type: %d\n", msg_length, msg_type);
#endif

      switch(msg_type) {
      case BGP_OPEN:
	break;
      case BGP_UPDATE:
#ifdef BGPDUMP_DEBUG
	fprintf(stdout, "UPDATE Message Received...\n");
#endif
	/* read length of withdrawn routes */
	memcpy((void*)&wr_len, (char*)(pkt+pkt_offset), 2);
	pkt_offset += 2;
	wr_len = ntohs(wr_len);
#ifdef BGPDUMP_DEBUG
	fprintf(stdout, "withdrawn routes length: %d\n", wr_len);
#endif

	if (wr_len) {
	  upd->ad_type = BGP_WITHDRAWAL;
	  tmpbuf_var = (char*)realloc((void*)tmpbuf_var, wr_len*sizeof(char));
	  memcpy((void*)tmpbuf_var, (char*)(pkt+pkt_offset), wr_len);
	  pkt_offset += wr_len;
	  /* read out withdrawn routes */
	  count = 0;

	  while (count < wr_len) {
	    upd->num_prefixes_wd++;
	    memcpy((void*)&pfxlen, (void*)(tmpbuf_var+count), 1*sizeof(char));
	    count++;
  	    pfxlen &= 0xff; 

	    memcpy((void*)&pfx, (void*)(tmpbuf_var+count),
		 (pfxlen/8 + ((pfxlen%8)?1:0))*sizeof(char));
	    pfx = ntohl(pfx);
	    pfx &= ((unsigned int)pow(2,pfxlen)-1) << (32-pfxlen);
	    pfxstr = (char*)intoa(pfx);

	    // WEIRD MEMORT SHIT REQUIRES THIS
  	    if (upd->prefix_==NULL || (upd->num_prefixes*sizeof(prefix_s) > sizeof(*(upd->prefix_))))
  	      upd->prefix_ = (prefix_s *)realloc((void*)(upd->prefix_), 2*upd->num_prefixes*sizeof(prefix_s));

	    if (upd->prefix_wd_==NULL || (upd->num_prefixes_wd*sizeof(prefix_s) > sizeof(*(upd->prefix_wd_))))
	      upd->prefix_wd_ = (prefix_s *)realloc((void*)(upd->prefix_wd_), 2*upd->num_prefixes_wd*sizeof(prefix_s));
    	    (upd->prefix_wd_)[upd->num_prefixes_wd-1].pfx = pfx;  
    	    (upd->prefix_wd_)[upd->num_prefixes_wd-1].len = pfxlen;  


	    
#ifdef BGPDUMP_DEBUG
	    fprintf(stdout, "prefix: 0x%.4x (%s), length: %d\n", pfx, pfxstr, pfxlen);
#endif
	    count+=(pfxlen/8) + ((pfxlen%8)?1:0);
	  }
	}

	/* path attribute length) */
	memcpy((void*)&pa_len, (char*)(pkt+pkt_offset), 2);
	pkt_offset += 2;

	pa_len = ntohs(pa_len);
#ifdef BGPDUMP_DEBUG
	fprintf(stdout, "path attr. length: %d\n", pa_len);
#endif
	if (pa_len) {
	  upd->ad_type = BGP_ADV;
	  tmpbuf_var = (char*)realloc((void*)tmpbuf_var, pa_len*sizeof(char));

	  memcpy((void*)tmpbuf_var, (char*)(pkt+pkt_offset), pa_len);
	  pkt_offset += pa_len;

	  /* read out all of the path attribute stuff */
	  count = 0;

	  while (count < pa_len) {

	    /* parse attribute flags */
	    memcpy((void*)&pa_fgs, (void*)(tmpbuf_var+count), 1*sizeof(char));
	    count++;

	    pa_fgs_str[0] = '\0';
#ifdef BGPDUMP_DEBUG
	    if (pa_fgs&0x80) {
	      strncat(pa_fgs_str, "O", 1);
	    }
	    if (pa_fgs&0x40) {
	      strncat(pa_fgs_str, "T", 1);
	    }
	    if (pa_fgs&0x20) {
	      strncat(pa_fgs_str, "P", 1);
	    }
	    if (pa_fgs&0x10) {
	      strncat(pa_fgs_str, "E", 1);
	    }
	    if (strlen(pa_fgs_str) > 0) { fprintf(stdout, "(%s)\t", pa_fgs_str); }
#endif

	    /* parse attribute type */
	    memcpy((void*)&pa_type, (void*)(tmpbuf_var+count), 1*sizeof(char));
	    count++;

	    /* parse attribute length */
	    if (pa_fgs&0x10) {
	      memcpy((void*)&attr_len, (void*)(tmpbuf_var+count), 2*sizeof(char));
	      attr_len = ntohs(attr_len);
	      count += 2;
	    } else {
	      memcpy((void*)&attr_len, (void*)(tmpbuf_var+count), 1*sizeof(char));
	      count += 1;
	      attr_len &= 0xff;
	    }  

	    /* parse attribute value */
	    as_path[0] = '\0';

	    switch(pa_type) {
	    case ORIGIN:
	      memcpy((void*)&origin, (tmpbuf_var+count), attr_len*sizeof(char));
	      origin &= 0x3;
	      upd->origin = origin;
#ifdef BGPDUMP_DEBUG
	      fprintf(stdout, "\tORIGIN: %d\n", origin);
#endif
	      break;
	    case AS_PATH:
	      memcpy((void*)&ps_type, (tmpbuf_var+count), 1*sizeof(char));
	      memcpy((void*)&ps_len, (tmpbuf_var+count+1), 1*sizeof(char));
	      for (i=0; i < 2*ps_len; i+=2) {
		memcpy((void*)&as_el, (tmpbuf_var+count+2+i), 2*sizeof(char));
		as_el = ntohs(as_el);
		sprintf(as_path, "%s%d ", as_path, as_el);
		if (strlen(as_path) > 80) {break;}
	      }
	      upd->origin_as = as_el;

	      as_path[strlen(as_path)-1] = '\0';
  	      strncpy(upd->as_path, as_path, strlen(as_path));
	      upd->as_path[strlen(as_path)] = '\0';
#ifdef BGPDUMP_DEBUG
	      fprintf(stdout, "\tASPATH: %s\n", as_path);
#endif
	      break;
	    case NEXT_HOP:
	      memcpy((void*)&next_hop, (tmpbuf_var+count), attr_len*sizeof(char));
	      next_hop = ntohl(next_hop);
	      next_hop_str = (char*)intoa(next_hop);
	      upd->next_hop = next_hop;
#ifdef BGPDUMP_DEBUG
	      fprintf(stdout, "\tNEXT HOP: %s\n", next_hop_str);
#endif
	      break;
	    case MULTI_EXIT_DISC:
  	      memcpy((void*)&med, (tmpbuf_var+count), attr_len*sizeof(char));
	      med = ntohl(med);
	      upd->med = med;
#ifdef BGPDUMP_DEBUG
	      fprintf(stdout, "\tMED: %d\n", med);
#endif
	      break;
	    case LOCAL_PREF:
	      memcpy((void*)&localpref, (tmpbuf_var+count), attr_len*sizeof(char));
	      localpref = ntohl(localpref);
	      upd->localpref = localpref;
#ifdef BGPDUMP_DEBUG
	      fprintf(stdout, "\tLOCALPREF: %d\n", localpref);
#endif
	      break;
	    case ATOMIC_AGGREGATE:
	      break;
	    case AGGREGATOR:
#ifdef BGPDUMP_DEBUG
	      fprintf(stdout, "\tAGGREGATOR: %s\n", "XXX");
#endif
	      break;
	    }
	    count += attr_len;
	  }

	  nlri_len = msg_length - 23 - pa_len - wr_len;
	  if (nlri_len && pa_len) {
	    /* read out advertised routes */ 
	    count=0;
	    tmpbuf_var = (char*)realloc((void*)tmpbuf_var, nlri_len*sizeof(char));
	    memcpy((void*)tmpbuf_var, (char*)(pkt+pkt_offset), nlri_len);
	    pkt_offset += nlri_len;

	    upd->num_prefixes++;
	    memcpy((void*)&pfxlen, (void*)(tmpbuf_var+count), 1*sizeof(char)); 
	    count++; 
	    pfxlen &= 0xff; 
		
	    memcpy((void*)&pfx, (void*)(tmpbuf_var+count),
		   (pfxlen/8 + ((pfxlen%8)?1:0))*sizeof(char));
	    count+=(pfxlen/8) + ((pfxlen%8)?1:0);

	    pfx = ntohl(pfx);
	    pfx &= ((unsigned int)pow(2,pfxlen)-1) << (32-pfxlen);
	    pfxstr = (char*)intoa(pfx);

	    if (upd->prefix_==NULL || (upd->num_prefixes*sizeof(prefix_s) > sizeof(*(upd->prefix_))))
	      upd->prefix_ = (prefix_s *)realloc((void*)(upd->prefix_), 2*upd->num_prefixes*sizeof(prefix_s));

	    
    	    (upd->prefix_)[upd->num_prefixes-1].pfx = pfx;  
    	    (upd->prefix_)[upd->num_prefixes-1].len = pfxlen;  
	    



#ifdef BGPDUMP_DEBUG
	    fprintf(stdout, "prefix: 0x%.4x (%s), length: %d\n", pfx, pfxstr, pfxlen);
#endif

	  } 
	}

	break;
      case BGP_NOTIFICATION:
	break;
      case BGP_KEEPALIVE:
	break;
      default: break;
      }

     
    case BGP4MP_ENTRY:
      break;

    case BGP4MP_SNAPSHOT:
      break;
    }
  default: break;
  }
  if (tmpbuf_var) { free(tmpbuf_var); }

  
}


void print_mrt_format(table_entry *te, FILE *fd) {
  printf("%s|%d|%c|%s|%d|%s/%d|%s|%s|%s|%d|%d||NAG||\n",
	 "TABLE_DUMP", te->uptime, 'B', te->peer_ip, te->peer_as,
	 te->tpfx, te->tlen, "<aspath>", "<origin>", "<next_hop>",
	 100, 1000);
}


void print_mrt_format(update *upd, FILE *fd) {
  
  char ad_char[2];
  char* orig_char[3];

  char dest_ip [sizeof(".xxx.xxx.xxx.xxx")];
  char next_hop[sizeof(".xxx.xxx.xxx.xxx")];
  char prefix[sizeof(".xxx.xxx.xxx.xxx/xx")];
  
  ad_char[BGP_ADV] = 'A';
  ad_char[BGP_WITHDRAWAL] = 'W';

  orig_char[IGP] = "IGP";
  orig_char[EGP] = "EGP";
  orig_char[INCOMPLETE] = "INCOMPLETE";

  strncpy(dest_ip, (char*)intoa(upd->next_hop), sizeof(dest_ip));
  strncpy(next_hop, (char*)intoa(upd->next_hop), sizeof(next_hop));

  for (int i=0; i<upd->num_prefixes; i++) {
    strncpy(prefix, (char*)intoa(upd->prefix_[i].pfx), sizeof(prefix));
    sprintf(prefix, "%s/%d", prefix, upd->prefix_[i].len);

//      if ((upd->ad_type & 1) == BGP_ADV) {
    // XXX add mask length to pfx
      // add dest IP to struct
      printf("%s|%d|%c|%s|%d|%s|%s|%s|%s|%d|%d||||\n",
	     "BGP4MP", upd->time, 'A',
	     next_hop, upd->src_as,
	     prefix, upd->as_path,
	     orig_char[(int)(upd->origin&0xff)], next_hop,
	     upd->localpref, upd->med);
      
//      } else if ((upd->ad_type & 1) == BGP_WITHDRAWAL) {
//        printf("%s|%d|%c|%s|%d|%s\n", "BGP4MP", upd->time, ad_char[upd->ad_type],
//  	     next_hop, upd->src_as, prefix);
//      }
  }


  for (int i=0; i<upd->num_prefixes_wd; i++) {
    strncpy(prefix, (char*)intoa(upd->prefix_wd_[i].pfx), sizeof(prefix));
    sprintf(prefix, "%s/%d", prefix, upd->prefix_wd_[i].len);
    printf("%s|%d|%c|%s|%d|%s\n", "BGP4MP", upd->time, 'W',
	   next_hop, upd->src_as, prefix);

  }

}



void print_db_format(update *upd, FILE *fd) {
  
  char ad_char[2];
  char orig_char[3];
  
  ad_char[BGP_ADV] = 'A';
  ad_char[BGP_WITHDRAWAL] = 'W';

  orig_char[IGP] = 'I';
  orig_char[EGP] = 'E';
  orig_char[INCOMPLETE] = '?';

  for (int i=0; i<upd->num_prefixes; i++) {

    printf("%d\t%c\t%d\t%d\t%lu\t%d\t%s\t%d\t%lu\t%d\t%d\t%c\n",
	     upd->time, 'A',
	     upd->src_as, upd->dst_as,
	     upd->prefix_[i].pfx, upd->prefix_[i].len,
	     upd->as_path, upd->origin_as,
	     upd->next_hop, upd->med, upd->localpref,
	     orig_char[(int)(upd->origin&0xff)]);
  }


  for (int i=0; i<upd->num_prefixes_wd; i++) {

    printf("%d\t%c\t%d\t%d\t%lu\t%d\t0\tNULL\t%lu\n",
	   upd->time, 'W',
	   upd->src_as, upd->dst_as,
	   upd->prefix_wd_[i].pfx, upd->prefix_wd_[i].len,
	   upd->next_hop);

  }

}






void print_long_format(update *upd, FILE *fd) {
  
  char ad_char[2];
  char* orig_char[3];

  char dest_ip [sizeof(".xxx.xxx.xxx.xxx")];
  char next_hop[sizeof(".xxx.xxx.xxx.xxx")];
  char prefix[sizeof(".xxx.xxx.xxx.xxx/xx")];
  
  ad_char[BGP_ADV] = 'A';
  ad_char[BGP_WITHDRAWAL] = 'W';

  orig_char[IGP] = "IGP";
  orig_char[EGP] = "EGP";
  orig_char[INCOMPLETE] = "INCOMPLETE";


  strncpy(dest_ip, (char*)intoa(upd->next_hop), sizeof(dest_ip));
  strncpy(next_hop, (char*)intoa(upd->next_hop), sizeof(next_hop));
  strncpy(prefix, (char*)intoa(upd->prefix_->pfx), sizeof(prefix));
  
  sprintf(prefix, "%s/%d", prefix, upd->prefix_->len);
  
  if ((upd->ad_type & 1) == BGP_ADV) {
    // XXX add mask length to pfx
    // add dest IP to struct
    printf("%s|%d|%c|%s|%d|%s|%s|%s|%s|%d|%d||||\n",
	    "BGP4MP", upd->time, ad_char[upd->ad_type],
	    next_hop, upd->src_as,
	    prefix, upd->as_path,
	    orig_char[(int)(upd->origin&0xff)], next_hop,
	    upd->localpref, upd->med);
	    
  } else if ((upd->ad_type & 1) == BGP_WITHDRAWAL) {
    printf("%s|%d|%c|%s|%d|%s\n", "BGP4MP", upd->time, ad_char[upd->ad_type],
	   next_hop, upd->src_as, prefix);
  }

}


void parse_file_mrt(char *filename, struct opts *opt, update_stats *stats) {

  FILE *fp;
  FILE *fstdout = NULL;//fopen("/dev/stdout", "w");
  update upd;
  table_entry te;
  te.peer_as=0;

  opt->all_files;

  unsigned long start_ep = opt->start_time_ep;
  unsigned long end_ep = opt->end_time_ep;


  fp = fopen(filename, "rb");
  if (fp == NULL) {
    fprintf(stderr, "error: could not open %s\n", filename);
    return;
  }
  
  

  while(!feof(fp)) {
    read_record(fp, &upd, &te);
    if (feof(fp)) {break;}
    if ((upd.time > start_ep && upd.time < end_ep) || (!start_ep)) {
      /* set by the -m option */
      if (opt->mrt_output) {
	
	if (upd.prefix_)
 	  print_db_format(&upd, fstdout);
//   	  print_mrt_format(&upd, fstdout);
	else if (te.peer_as) {
  	  print_mrt_format(&te, fstdout);
	}

      }
      if (opt->do_stats) {
  	stats->update(&upd);
      }
	      
    }
    
//      if (upd.prefix_) { free(upd.prefix_); upd.prefix_=NULL; }
//      if (upd.prefix) { free(upd.prefix); upd.prefix=NULL; }

  }
  fclose(fp);
//  fclose(fstdout);
}

void read_update_file(char *filename) {

  FILE *fp;
  update upd;
  table_entry te;
 
  fp = fopen(filename, "rb");
  if (fp == NULL) { fprintf(stderr, "error: could not open %s\n", filename); };
  
  while(!feof(fp)) {
    read_record(fp, &upd, &te);
  }
  fclose(fp);
}
    
    
