/*
 * Copyright (c) 2001 Massachusetts Institute of Technology.
 *
 * This file is part of bgptools. 
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `COPYING' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file AUTHORS which should have accompanied this software 
 * for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the AUTHORS file.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


/*
 *     Nick Feamster (feamster@lcs.mit.edu)
 *     MIT Lab for Computer Science
 *     November 2001
 */

#include "trace.h"

void _accept_cb(ClientData cd, int mask) {
  Trace *tp = (Trace*)cd;
  tp->accept_cb();
}

void _read_cb(ClientData cd, int mask) {
  Trace *tp = (Trace*)cd;
  tp->read_cb();
}


Trace::Trace(char *pathname, char *tracefile) {

  this->tracefile = tracefile;

#ifdef BGPDUMP_DEBUG
  fprintf(stderr, "UNIX domain socket on %s\n", pathname);
#endif
  if ((sock = socket(AF_LOCAL, SOCK_STREAM, 0)) == -1) {
    perror("::Trace:socket()");
    return;
  }
  nonblock(sock);
  nonblock(csock);
  
  unlink(pathname);
  bzero(&servaddr, sizeof(servaddr));

  servaddr.sun_family = AF_LOCAL;
  strncpy(servaddr.sun_path, pathname, sizeof(servaddr.sun_path)-1);

  if ((bind(sock, (struct sockaddr*)&servaddr, sizeof(servaddr))) == -1) {
    perror("::Trace:bind()");
    return;
  }

  if ((listen(sock, 10)) == -1) {
    perror("::Trace:listen()");
    return;
  }

  Tcl_CreateFileHandler(sock, TCL_READABLE,
    			(Tcl_FileProc*)&(_accept_cb), this);
  
}

void
Trace::accept_cb() {
#ifdef BGPDUMP_DEBUG
  fprintf(stderr, "Trace::accept_cb: accepting connections...\n");
#endif
  socklen_t clilen = sizeof(struct sockaddr_un);
  if ((csock = accept(sock, (sockaddr*)&cliaddr, &clilen)) < 0) {
    perror("Trace::accept");
    return;
  }

  Tcl_CreateFileHandler(csock, TCL_READABLE,
			(Tcl_FileProc*)&(_read_cb), this);
}

void
Trace::read_cb() {
#ifdef BGPDUMP_DEBUG
  fprintf(stderr, "Trace::read_cb: reading...\n");
#endif
  int n;
  if ((n=read(csock, (void*)&inbuf, BUF_SZ)) < 0) {
    perror("Trace::read_cb");
    return;
  }
  if (!n) {
    Tcl_DeleteFileHandler(csock);
    close(csock);
    return;
  }
  inbuf[n]='\0';
#ifdef BGPDUMP_DEBUG
  fprintf(stderr, "--got %d bytes\n[%s]\n", n, inbuf);
#endif
  run_trace(inbuf, n);

}

void
Trace::run_trace(char *prefix, int len) {
  char cmd[256];
  char pfx[sizeof(".xxx.xxx.xxx.xxx/xx")];
  char pfx_m[sizeof(".xxx.xxx.xxx.xxx/xx")];
  FILE *tracef;
  pid_t pid;


  if (!(strcmp(prefix, "::"))) {return;}

  if (strstr(prefix, "::")) {
    char *tmp = (char*)malloc(strlen(prefix));
    strncpy(tmp, prefix, len);
    while (1) {
      char *tok = strsep(&tmp, "::");
      if (!tok) {break;}
      if (!strlen(tok)) {continue;}
      
      strncpy(pfx_m, tok, strlen(tok));
      pfx_m[strlen(tok)]='\0';

      char *slash = strstr(tok, "/");
      if (slash) {*slash = '\0';}

      strncpy(pfx, tok, strlen(tok));
      pfx[strlen(tok)-1] = '1';
      pfx[strlen(tok)] = '\0';
	sprintf(cmd, "%s -- %s %s %s %s >> %s",
		"/usr/local/bin/bufwrite", "/usr/local/bin/traceroute", pfx_m,
		pfx, "-n", tracefile);
#ifdef BGPDUMP_DEBUG
	fprintf(stderr, "running \"%s\" \n", cmd);
#endif
  	if ((num_children < MAX_TRACEROUTE) && !(pid = fork())) {
	  system(cmd);
  	  _exit(0);
  	} else {
	  num_children++;
	}
      
    }
    free(tmp);
  } else {
    strncpy(pfx_m, prefix, len);
    pfx_m[len]='\0';

      char *slash = strstr(prefix, "/");
      if (slash) {*slash = '\0';}
    
      prefix[strlen(prefix)-1] = '1';
      sprintf(cmd, "%s -- %s %s %s %s >> %s",
	      "/usr/local/bin/bufwrite", "/usr/local/bin/traceroute", pfx_m,
	      prefix, "-n", tracefile);
#ifdef BGPDUMP_DEBUG
    fprintf(stderr, "running \"%s\" \n", cmd);
#endif

    if ((num_children < MAX_TRACEROUTE) && !(pid = fork())) {
      system(cmd);
      _exit(0);
    } else {
      num_children++;
    }
  }

  while (num_children > 0) {
    wait(0);
    num_children--;
  }
    

}

Trace::~Trace() {
}


void
Trace::nonblock(int fd)
{       
#ifdef WIN32
        u_long flag = 1;
        if (ioctlsocket(fd, FIONBIO, &flag) == -1) {
                fprintf(stderr, "ioctlsocket: FIONBIO: %lu\n", GetLastError());
                exit(1);
        }
#else
        int flags = fcntl(fd, F_GETFL, 0);
#if defined(hpux) || defined(__hpux)
        flags |= O_NONBLOCK;
#else
        flags |= O_NONBLOCK|O_NDELAY;
#endif
        if (fcntl(fd, F_SETFL, flags) == -1) {
                perror("fcntl: F_SETFL");
                exit(1);
        }
#endif
}


/****************************************************************************/

traceClient::traceClient(char *pathname) {
  
  if ((csock = socket(AF_LOCAL, SOCK_STREAM, 0)) < 0)
    perror("traceClient::socket()");

  int flag = 1;
  int result = setsockopt(csock, IPPROTO_TCP,
			  TCP_NODELAY,
			  (char *) &flag,
			  sizeof(int));
  
  bzero(&servaddr, sizeof(struct sockaddr_un));
  servaddr.sun_family = AF_LOCAL;
  strcpy(servaddr.sun_path, pathname);
}

int
traceClient::connect_() {
  int ret;
  if ((ret = connect(csock,
		     (sockaddr *)&servaddr, sizeof(struct sockaddr_un))) < 0)
    perror("traceClient::connect()");
  return ret;
}

int
traceClient::close_() {
  return close(csock);
}

int
traceClient::sendUpdates(char *filename) {
  FILE *fp = fopen(filename, "r");
  update upd;
  char prefix[sizeof(".xxx.xxx.xxx.xxx/xx")];
  int ret=0;
  
  if (fp == NULL) {
    fprintf(stderr, "error: could not open %s\n", filename);
    return -1;
  }

  while(!feof(fp)) {
    read_record(fp, &upd);
    for (int i=0; i<upd.num_prefixes_wd; i++) {
      char *pstr = (char*)intoa(upd.prefix_wd_[i].pfx);
      strncpy(prefix, pstr, strlen(pstr));
      sprintf(prefix, "%s/%d%c", pstr, upd.prefix_wd_[i].len, '\0');
      ret = this->send(prefix,strlen(prefix));
      this->send("::", 2);
    }
  }
  fclose(fp);
  return ret;
}

int
traceClient::sendUpdate(struct update *upd) {
  char prefix[sizeof(".xxx.xxx.xxx.xxx/xx")];
  int ret=0;

      for (int i=0; i<upd->num_prefixes_wd; i++) {
	char *pstr = (char*)intoa(upd->prefix_wd_[i].pfx);
	strncpy(prefix, pstr, strlen(pstr));
  	sprintf(prefix, "%s/%d%c", pstr, upd->prefix_wd_[i].len, '\0');
	ret = this->send(prefix,strlen(prefix));
	this->send("::", 2);
      }
  return ret;
}

int
traceClient::sendUpdate(char *pkt, int plen) {
  char prefix[sizeof(".xxx.xxx.xxx.xxx/xx")];
  update upd;
  int ret=0;

  read_record(pkt, plen, &upd);
  
  for (int i=0; i<upd.num_prefixes_wd; i++) {
    char *pstr = (char*)intoa(upd.prefix_wd_[i].pfx);
    strncpy(prefix, pstr, strlen(pstr));
    sprintf(prefix, "%s/%d%c", pstr, upd.prefix_wd_[i].len, '\0');
    ret = this->send(prefix,strlen(prefix));
    this->send("::", 2);
  }

  return ret;
}




int
traceClient::send(char *buf, int len) {
  int n;
  if ((n =write(csock, buf,len)) < 0) {
    perror("traceClient::write()");
    return -1;
  }
  return n;
}
  
traceClient::~traceClient() {
  close(csock);
}



//      tracef = popen(cmd, "r");
//      char tmp[80];
//      while (!feof(tracef)) {
//        fread((void*)tmp, sizeof(char), 16, tracef);
//        tmp[16] = '\0';
//        fprintf(trace_fp, "%s", tmp);
//        if (strstr(tmp, "* * *")) { break; }
//      }
